@@ -120,6 +120,13 @@ static PGLogicalRelation   *last_insert_rel = NULL;
 static int					last_insert_rel_cnt = 0;
 static bool					use_multi_insert = false;
 
+/*
+ * A message counter for the xact, for debugging. We don't send
+ * the remote change LSN with messages, so this aids identification
+ * of which change causes an error.
+ */
+static uint32			xact_action_counter;
+
 typedef struct PGLFlushPosition
 {
 	dlist_node node;
@@ -137,13 +144,24 @@ typedef struct ApplyExecState
 	TupleTableSlot	   *slot;
 } ApplyExecState;
 
+struct ActionErrCallbackArg
+{
+	const char * action_name;
+	PGLogicalRelation *rel;
+	bool is_ddl_or_drop;
+	bool suppress_output;
+};
+
+struct ActionErrCallbackArg errcallback_arg;
+static TransactionId remote_xid;
+
 static void multi_insert_finish(void);
 
 static void handle_queued_message(HeapTuple msgtup, bool tx_just_started);
 static void handle_startup_param(const char *key, const char *value);
 static bool parse_bool_param(const char *key, const char *value);
 static void process_syncing_tables(XLogRecPtr end_lsn);
-static void start_sync_worker(RangeVar *rv);
+static void start_sync_worker(Name nspname, Name relname);
 
 /*
  * Check if given relation is in process of being synchronized.
@@ -151,10 +169,94 @@ static void start_sync_worker(RangeVar *rv);
  * TODO: performance
  */
 static bool
-check_syncing_relation(const char *nspname, const char *relname)
+should_apply_changes_for_rel(const char *nspname, const char *relname)
 {
-	return list_length(SyncingTables) &&
-		list_member(SyncingTables, makeRangeVar((char *)nspname, (char *)relname, -1));
+	if (list_length(SyncingTables) > 0)
+	{
+		ListCell	   *lc;
+
+		foreach (lc, SyncingTables)
+		{
+			PGLogicalSyncStatus	   *sync = (PGLogicalSyncStatus *) lfirst(lc);
+
+			if (namestrcmp(&sync->nspname, nspname) == 0 &&
+				namestrcmp(&sync->relname, relname) == 0 &&
+				(sync->status != SYNC_STATUS_READY &&
+				 !(sync->status == SYNC_STATUS_SYNCDONE &&
+				   sync->statuslsn <= replorigin_session_origin_lsn)))
+				return false;
+		}
+	}
+
+	return true;
+}
+
+/*
+ * Prepare apply state details for errcontext or direct logging.
+ *
+ * This callback could be invoked at all sorts of weird times
+ * so it should assume as little as psosible about the invoking
+ * context.
+ */
+static void
+format_action_description(
+	StringInfo si,
+	const char * action_name,
+	PGLogicalRelation *rel,
+	bool is_ddl_or_drop)
+{
+	appendStringInfoString(si, "apply ");
+	appendStringInfoString(si,
+		action_name == NULL ? "(unknown action)" : action_name);
+
+	if (rel != NULL && 
+		rel->nspname != NULL
+		&& rel->relname != NULL
+		&& !is_ddl_or_drop)
+	{
+		appendStringInfo(si, " from remote relation %s.%s",
+				rel->nspname, rel->relname);
+	}
+
+	appendStringInfo(si,
+			" in commit before %X/%X, xid %u commited at %s (action #%u)",
+			(uint32)(replorigin_session_origin_lsn>>32),
+			(uint32)replorigin_session_origin_lsn,
+			remote_xid,
+			timestamptz_to_str(replorigin_session_origin_timestamp),
+			xact_action_counter);
+
+	if (replorigin_session_origin != InvalidRepOriginId)
+	{
+		appendStringInfo(si, " from node replorigin %u",
+			replorigin_session_origin);
+	}
+
+	if (remote_origin_id != InvalidRepOriginId)
+	{
+		appendStringInfo(si, " forwarded from commit %X/%X on node %u",
+				(uint32)(remote_origin_lsn>>32),
+				(uint32)remote_origin_lsn,
+				remote_origin_id);
+	}
+}
+
+static void
+action_error_callback(void *arg)
+{
+	StringInfoData si;
+
+	if (!errcallback_arg.suppress_output)
+	{
+		initStringInfo(&si);
+
+		format_action_description(&si,
+			errcallback_arg.action_name,
+			errcallback_arg.rel,
+			errcallback_arg.is_ddl_or_drop);
+
+		errcontext("%s", si.data);
+	}
 }
 
 static bool
@@ -167,6 +269,15 @@ ensure_transaction(void)
 		return false;
 	}
 
+	/*
+	 * pglogical doesn't have "statements" as such, so we'll report one
+	 * statement per applied transaction. We must set the statement start time
+	 * because StartTransaction() uses it to initialize the transaction cached
+	 * timestamp used by current_timestamp. If we don't set it, every xact will
+	 * get the same current_timestamp. See 2ndQuadrant/pglogical_internal#148
+	 */
+	SetCurrentStatementStartTimestamp();
+
 	StartTransactionCommand();
 	apply_api.on_begin();
 	MemoryContextSwitchTo(MessageContext);
@@ -179,7 +290,9 @@ handle_begin(StringInfo s)
 {
 	XLogRecPtr		commit_lsn;
 	TimestampTz		commit_time;
-	TransactionId	remote_xid;
+
+	xact_action_counter = 1;
+	errcallback_arg.action_name = "BEGIN";
 
 	pglogical_read_begin(s, &commit_lsn, &commit_time, &remote_xid);
 
@@ -224,6 +337,9 @@ handle_commit(StringInfo s)
 	XLogRecPtr		end_lsn;
 	TimestampTz		commit_time;
 
+	errcallback_arg.action_name = "COMMIT";
+	xact_action_counter++;
+
 	pglogical_read_commit(s, &commit_lsn, &end_lsn, &commit_time);
 
 	Assert(commit_time == replorigin_session_origin_timestamp);
@@ -281,12 +397,21 @@ handle_commit(StringInfo s)
 	if (remote_origin_id != InvalidRepOriginId &&
 		remote_origin_id != replorigin_session_origin)
 	{
+#if PG_VERSION_NUM >= 90500
+		Relation replorigin_rel;
+#endif
 		elog(DEBUG3, "advancing origin oid %u for forwarded row to %X/%X",
 			remote_origin_id,
 			(uint32)(XactLastCommitEnd>>32), (uint32)XactLastCommitEnd);
 
+#if PG_VERSION_NUM >= 90500
+		replorigin_rel = heap_open(ReplicationOriginRelationId, RowExclusiveLock);
+#endif
 		replorigin_advance(remote_origin_id, remote_origin_lsn,
 						   XactLastCommitEnd, false, false /* XXX ? */);
+#if PG_VERSION_NUM >= 90500
+		heap_close(replorigin_rel, RowExclusiveLock);
+#endif
 	}
 
 	in_remote_transaction = false;
@@ -305,6 +430,19 @@ handle_commit(StringInfo s)
 				 (uint32)(MyApplyWorker->replay_stop_lsn >>32),
 				 (uint32)MyApplyWorker->replay_stop_lsn)));
 
+		/*
+		 * If this is sync worker, update syncing table state to done.
+		 */
+		if (MyPGLogicalWorker->worker_type == PGLOGICAL_WORKER_SYNC)
+		{
+			StartTransactionCommand();
+			set_table_sync_status(MyApplyWorker->subid,
+							  NameStr(MyPGLogicalWorker->worker.sync.nspname),
+							  NameStr(MyPGLogicalWorker->worker.sync.relname),
+							  SYNC_STATUS_SYNCDONE, end_lsn);
+			CommitTransactionCommand();
+		}
+
 		/*
 		 * Flush all writes so the latest position can be reported back to the
 		 * sender.
@@ -329,6 +467,9 @@ handle_commit(StringInfo s)
 		proc_exit(0);
 	}
 
+	xact_action_counter = 0;
+	remote_xid = InvalidTransactionId;
+
 	process_syncing_tables(end_lsn);
 
 	pgstat_report_activity(STATE_IDLE, NULL);
@@ -377,10 +518,14 @@ handle_insert(StringInfo s)
 	PGLogicalRelation  *rel;
 	bool				started_tx = ensure_transaction();
 
+	errcallback_arg.action_name = "INSERT";
+	xact_action_counter++;
+
 	rel = pglogical_read_insert(s, RowExclusiveLock, &newtup);
+	errcallback_arg.rel = rel;
 
 	/* If in list of relations which are being synchronized, skip. */
-	if (check_syncing_relation(rel->nspname, rel->relname))
+	if (!should_apply_changes_for_rel(rel->nspname, rel->relname))
 	{
 		pglogical_relation_close(rel, NoLock);
 		return;
@@ -465,11 +610,19 @@ multi_insert_finish(void)
 {
 	if (use_multi_insert && last_insert_rel_cnt)
 	{
+		const char *old_action = errcallback_arg.action_name;
+		PGLogicalRelation *old_rel = errcallback_arg.rel;
+		errcallback_arg.action_name = "multi INSERT";
+		errcallback_arg.rel = last_insert_rel;
+
 		apply_api.multi_insert_finish(last_insert_rel);
 		pglogical_relation_close(last_insert_rel, NoLock);
 		use_multi_insert = false;
 		last_insert_rel = NULL;
 		last_insert_rel_cnt = 0;
+
+		errcallback_arg.rel = old_rel;
+		errcallback_arg.action_name = old_action;
 	}
 }
 
@@ -481,15 +634,19 @@ handle_update(StringInfo s)
 	PGLogicalRelation  *rel;
 	bool				hasoldtup;
 
+	errcallback_arg.action_name = "UPDATE";
+	xact_action_counter++;
+
 	ensure_transaction();
 
 	multi_insert_finish();
 
 	rel = pglogical_read_update(s, RowExclusiveLock, &hasoldtup, &oldtup,
 								&newtup);
+	errcallback_arg.rel = rel;
 
 	/* If in list of relations which are being synchronized, skip. */
-	if (check_syncing_relation(rel->nspname, rel->relname))
+	if (!should_apply_changes_for_rel(rel->nspname, rel->relname))
 	{
 		pglogical_relation_close(rel, NoLock);
 		return;
@@ -506,14 +663,18 @@ handle_delete(StringInfo s)
 	PGLogicalTupleData	oldtup;
 	PGLogicalRelation  *rel;
 
+	memset(&errcallback_arg, 0, sizeof(struct ActionErrCallbackArg));
+	xact_action_counter++;
+
 	ensure_transaction();
 
 	multi_insert_finish();
 
 	rel = pglogical_read_delete(s, RowExclusiveLock, &oldtup);
+	errcallback_arg.rel = rel;
 
 	/* If in list of relations which are being synchronized, skip. */
-	if (check_syncing_relation(rel->nspname, rel->relname))
+	if (!should_apply_changes_for_rel(rel->nspname, rel->relname))
 	{
 		pglogical_relation_close(rel, NoLock);
 		return;
@@ -700,7 +861,7 @@ handle_truncate(QueuedMessage *queued_message)
 	rv = parse_relation_message(queued_message->message);
 
 	/* If in list of relations which are being synchronized, skip. */
-	if (check_syncing_relation(rv->schemaname, rv->relname))
+	if (!should_apply_changes_for_rel(rv->schemaname, rv->relname))
 		return;
 
 	truncate_table(rv->schemaname, rv->relname);
@@ -712,10 +873,10 @@ handle_truncate(QueuedMessage *queued_message)
 static void
 handle_table_sync(QueuedMessage *queued_message)
 {
-	RangeVar	   *rv;
+	RangeVar			   *rv;
 	MemoryContext			oldcontext;
-	PGLogicalSyncStatus		*oldsync;
-	PGLogicalSyncStatus		newsync;
+	PGLogicalSyncStatus	   *oldsync;
+	PGLogicalSyncStatus	   *newsync;
 
 	rv = parse_relation_message(queued_message->message);
 
@@ -731,22 +892,20 @@ handle_table_sync(QueuedMessage *queued_message)
 		return;
 	}
 
-	/*
-	 * Create synchronization status for the subscription.
-	 * Currently we only support data sync for tables.
-	 */
-	newsync.kind = SYNC_KIND_DATA;
-	newsync.subid = MyApplyWorker->subid;
-	newsync.nspname = rv->schemaname;
-	newsync.relname = rv->relname;
-	newsync.status = SYNC_STATUS_INIT;
-	create_local_sync_status(&newsync);
-
 	/* Keep the lists persistent. */
 	oldcontext = MemoryContextSwitchTo(TopMemoryContext);
-	SyncingTables = lappend(SyncingTables,
-							makeRangeVar(pstrdup(rv->schemaname),
-										 pstrdup(rv->relname), -1));
+	newsync = palloc0(sizeof(PGLogicalSyncStatus));
+	MemoryContextSwitchTo(oldcontext);
+
+	newsync->kind = SYNC_KIND_DATA;
+	newsync->subid = MyApplyWorker->subid;
+	newsync->status = SYNC_STATUS_INIT;
+	namestrcpy(&newsync->nspname, rv->schemaname);
+	namestrcpy(&newsync->relname, rv->relname);
+	create_local_sync_status(newsync);
+
+	oldcontext = MemoryContextSwitchTo(TopMemoryContext);
+	SyncingTables = lappend(SyncingTables, newsync);
 	MemoryContextSwitchTo(oldcontext);
 }
 
@@ -886,33 +1045,53 @@ handle_sql(QueuedMessage *queued_message, bool tx_just_started)
 static void
 handle_queued_message(HeapTuple msgtup, bool tx_just_started)
 {
-	QueuedMessage  *queued_message = queued_message_from_tuple(msgtup);
+	QueuedMessage  *queued_message;
+	const char	   *old_action_name;
+
+	old_action_name = errcallback_arg.action_name;
+	errcallback_arg.is_ddl_or_drop = true;
+
+	queued_message = queued_message_from_tuple(msgtup);
 
 	switch (queued_message->message_type)
 	{
 		case QUEUE_COMMAND_TYPE_SQL:
+			errcallback_arg.action_name = "QUEUED_SQL";
 			handle_sql(queued_message, tx_just_started);
 			break;
 		case QUEUE_COMMAND_TYPE_TRUNCATE:
+			errcallback_arg.action_name = "QUEUED_TRUNCATE";
 			handle_truncate(queued_message);
 			break;
 		case QUEUE_COMMAND_TYPE_TABLESYNC:
+			errcallback_arg.action_name = "QUEUED_TABLESYNC";
 			handle_table_sync(queued_message);
 			break;
 		case QUEUE_COMMAND_TYPE_SEQUENCE:
+			errcallback_arg.action_name = "QUEUED_SEQUENCE";
 			handle_sequence(queued_message);
 			break;
 		default:
 			elog(ERROR, "unknown message type '%c'",
 				 queued_message->message_type);
 	}
+
+	errcallback_arg.action_name = old_action_name;
+	errcallback_arg.is_ddl_or_drop = false;
 }
 
 static void
 replication_handler(StringInfo s)
 {
+	ErrorContextCallback errcallback;
 	char action = pq_getmsgbyte(s);
 
+	memset(&errcallback_arg, 0, sizeof(struct ActionErrCallbackArg));
+	errcallback.callback = action_error_callback;
+	errcallback.arg = &errcallback_arg;
+	errcallback.previous = error_context_stack;
+	error_context_stack = &errcallback;
+
 	switch (action)
 	{
 		/* BEGIN */
@@ -950,6 +1129,9 @@ replication_handler(StringInfo s)
 		default:
 			elog(ERROR, "unknown action of type %c", action);
 	}
+
+	if (error_context_stack == &errcallback)
+		error_context_stack = errcallback.previous;
 }
 
 /*
@@ -1264,6 +1446,12 @@ pglogical_execute_sql_command(char *cmdstr, char *role, bool isTopLevel)
 	errcallback.previous = error_context_stack;
 	error_context_stack = &errcallback;
 
+
+	/*
+	 * XL distributes individual statements using just executing them as plain
+	 * SQL query and can't handle multistatements this way so we need to get
+	 * individual statements using API provided by XL itself.
+	 */
 #ifdef PGXC
 	commands = pg_parse_query_get_source(cmdstr, &commandSourceQueries);
 #else
@@ -1287,15 +1475,17 @@ pglogical_execute_sql_command(char *cmdstr, char *role, bool isTopLevel)
 	{
 		List	   *plantree_list;
 		List	   *querytree_list;
-		Node	   *command = (Node *) lfirst(command_i);
-#ifdef PGXC
-		char	   *commandSource = (char *) lfirst(commandSourceQuery_i);
-#endif
+		RawStmt	   *command = (RawStmt *) lfirst(command_i);
 		const char *commandTag;
 		Portal		portal;
 		int			save_nestlevel;
 		DestReceiver *receiver;
 
+#ifdef PGXC
+		cmdstr = (char *) lfirst(commandSourceQuery_i);
+		errcallback.arg = cmdstr;
+#endif
+
 		/* temporarily push snapshot for parse analysis/planning */
 		PushActiveSnapshot(GetTransactionSnapshot());
 
@@ -1312,11 +1502,7 @@ pglogical_execute_sql_command(char *cmdstr, char *role, bool isTopLevel)
 
 		querytree_list = pg_analyze_and_rewrite(
 			command,
-#ifdef PGXC
-			commandSource,
-#else
 			cmdstr,
-#endif
 			NULL, 0);
 
 		plantree_list = pg_plan_queries(
@@ -1326,11 +1512,7 @@ pglogical_execute_sql_command(char *cmdstr, char *role, bool isTopLevel)
 
 		portal = CreatePortal("pglogical", true, true);
 		PortalDefineQuery(portal, NULL,
-#ifdef PGXC
-						  commandSource,
-#else
 						  cmdstr,
-#endif
 						  commandTag,
 						  plantree_list, NULL);
 		PortalStart(portal, NULL, 0, InvalidSnapshot);
@@ -1379,19 +1561,16 @@ reread_unsynced_tables(Oid subid)
 
 		for (lc = list_head(SyncingTables); lc; lc = next)
 		{
-			RangeVar	   *rv = (RangeVar *) lfirst(lc);
+			PGLogicalSyncStatus	   *sync = (PGLogicalSyncStatus *) lfirst(lc);
 
 			next = lnext(lc);
 
-			pfree(rv->schemaname);
-			pfree(rv->relname);
-			pfree(rv);
+			pfree(sync);
 			pfree(lc);
 		}
 
 		pfree(SyncingTables);
 		SyncingTables = NIL;
-
 	}
 
 	/* Read new state. */
@@ -1399,10 +1578,9 @@ reread_unsynced_tables(Oid subid)
 	saved_ctx = MemoryContextSwitchTo(TopMemoryContext);
 	foreach (lc, unsynced_tables)
 	{
-		RangeVar	   *rv = lfirst(lc);
-		SyncingTables = lappend(SyncingTables,
-								makeRangeVar(pstrdup(rv->schemaname),
-											 pstrdup(rv->relname), -1));
+		PGLogicalSyncStatus	   *sync = palloc(sizeof(PGLogicalSyncStatus));
+		memcpy(sync, lfirst(lc), sizeof(PGLogicalSyncStatus));
+		SyncingTables = lappend(SyncingTables, sync);
 	}
 
 	MemoryContextSwitchTo(saved_ctx);
@@ -1411,6 +1589,8 @@ reread_unsynced_tables(Oid subid)
 static void
 process_syncing_tables(XLogRecPtr end_lsn)
 {
+	ListCell	   *lc;
+
 	/* First check if we need to update the cached information. */
 	if (MyApplyWorker->sync_pending)
 	{
@@ -1423,51 +1603,63 @@ process_syncing_tables(XLogRecPtr end_lsn)
 	/* Process currently pending sync tables. */
 	if (list_length(SyncingTables) > 0)
 	{
-		ListCell	   *lc,
-					   *prev,
+		ListCell	   *prev,
 					   *next;
 
 		prev = NULL;
 		for (lc = list_head(SyncingTables); lc; lc = next)
 		{
-			RangeVar	   *rv = (RangeVar *) lfirst(lc);
-			PGLogicalSyncStatus	   *sync;
-			char					status;
+			PGLogicalSyncStatus	   *sync = (PGLogicalSyncStatus *) lfirst(lc);
+			PGLogicalSyncStatus	   *newsync;
 
 			/* We might delete the cell so advance it now. */
 			next = lnext(lc);
 
 			StartTransactionCommand();
-			sync = get_table_sync_status(MyApplyWorker->subid, rv->schemaname,
-										 rv->relname, true);
+			newsync = get_table_sync_status(MyApplyWorker->subid,
+										 NameStr(sync->nspname),
+										 NameStr(sync->relname), true);
 
 			/*
 			 * TODO: what to do here? We don't really want to die,
 			 * but this can mean many things, for now we just assume table is
 			 * not relevant for us anymore and leave fixing to the user.
+			 *
+			 * The reason why this part happens in transaction is that the
+			 * memory allocated for sync info will get automatically cleaned
+			 * afterwards.
 			 */
-			if (!sync)
-				status = SYNC_STATUS_READY;
+			if (!newsync)
+			{
+				sync->status = SYNC_STATUS_READY;
+				sync->statuslsn = InvalidXLogRecPtr;
+			}
 			else
-				status = sync->status;
+				memcpy(sync, newsync, sizeof(PGLogicalSyncStatus));
 			CommitTransactionCommand();
 
-			if (status == SYNC_STATUS_SYNCWAIT)
+			if (sync->status == SYNC_STATUS_SYNCWAIT)
 			{
 				PGLogicalWorker *worker;
 
 				LWLockAcquire(PGLogicalCtx->lock, LW_EXCLUSIVE);
 				worker = pglogical_sync_find(MyDatabaseId,
 											 MyApplyWorker->subid,
-											 rv->schemaname, rv->relname);
+											 NameStr(sync->nspname),
+											 NameStr(sync->relname));
+
 				if (pglogical_worker_running(worker) &&
 					end_lsn >= worker->worker.apply.replay_stop_lsn)
 				{
 					worker->worker.apply.replay_stop_lsn = end_lsn;
+					sync->status = SYNC_STATUS_CATCHUP;
 
 					StartTransactionCommand();
-					set_table_sync_status(MyApplyWorker->subid, rv->schemaname,
-										  rv->relname, SYNC_STATUS_CATCHUP);
+					set_table_sync_status(MyApplyWorker->subid,
+										  NameStr(sync->nspname),
+										  NameStr(sync->relname),
+										  sync->status,
+										  sync->statuslsn);
 					CommitTransactionCommand();
 
 					if (pglogical_worker_running(worker))
@@ -1475,22 +1667,36 @@ process_syncing_tables(XLogRecPtr end_lsn)
 					LWLockRelease(PGLogicalCtx->lock);
 
 					if (wait_for_sync_status_change(MyApplyWorker->subid,
-													rv->schemaname,
-													rv->relname,
-													SYNC_STATUS_READY))
-						status = SYNC_STATUS_READY;
+													NameStr(sync->nspname),
+													NameStr(sync->relname),
+													SYNC_STATUS_SYNCDONE,
+													&sync->statuslsn))
+						sync->status = SYNC_STATUS_SYNCDONE;
 				}
 				else
 					LWLockRelease(PGLogicalCtx->lock);
 			}
 
+			if (sync->status == SYNC_STATUS_SYNCDONE &&
+				end_lsn >= sync->statuslsn)
+			{
+				sync->status = SYNC_STATUS_READY;
+				sync->statuslsn = end_lsn;
+
+				StartTransactionCommand();
+				set_table_sync_status(MyApplyWorker->subid,
+									  NameStr(sync->nspname),
+									  NameStr(sync->relname),
+									  sync->status,
+									  sync->statuslsn);
+				CommitTransactionCommand();
+			}
+
 			/* Ready? Remove it from local cache. */
-			if (status == SYNC_STATUS_READY)
+			if (sync->status == SYNC_STATUS_READY)
 			{
 				SyncingTables = list_delete_cell(SyncingTables, lc, prev);
-				pfree(rv->schemaname);
-				pfree(rv->relname);
-				pfree(rv);
+				pfree(sync);
 			}
 			else
 				prev = lc;
@@ -1501,17 +1707,21 @@ process_syncing_tables(XLogRecPtr end_lsn)
 	 * If there are still pending tables for syncrhonization, launch the sync
 	 * worker.
 	 */
-	if (list_length(SyncingTables) > 0)
+	foreach (lc, SyncingTables)
 	{
 		List		   *workers;
-		ListCell	   *lc;
+		ListCell	   *wlc;
 		int				nworkers = 0;
+		PGLogicalSyncStatus	   *sync = (PGLogicalSyncStatus *) lfirst(lc);
+
+		if (sync->status == SYNC_STATUS_SYNCDONE || sync->status == SYNC_STATUS_READY)
+			continue;
 
 		LWLockAcquire(PGLogicalCtx->lock, LW_EXCLUSIVE);
 		workers = pglogical_sync_find_all(MyDatabaseId, MyApplyWorker->subid);
-		foreach (lc, workers)
+		foreach (wlc, workers)
 		{
-			PGLogicalWorker	   *worker = (PGLogicalWorker *) lfirst(lc);
+			PGLogicalWorker	   *worker = (PGLogicalWorker *) lfirst(wlc);
 
 			if (pglogical_worker_running(worker))
 				nworkers++;
@@ -1520,15 +1730,14 @@ process_syncing_tables(XLogRecPtr end_lsn)
 
 		if (nworkers < 1)
 		{
-			RangeVar	   *rv = linitial(SyncingTables);
-
-			start_sync_worker(rv);
+			start_sync_worker(&sync->nspname, &sync->relname);
+			break;
 		}
 	}
 }
 
 static void
-start_sync_worker(RangeVar *rv)
+start_sync_worker(Name nspname, Name relname)
 {
 	PGLogicalWorker			worker;
 
@@ -1541,8 +1750,8 @@ start_sync_worker(RangeVar *rv)
 
 	/* Tell the worker to stop at current position. */
 	worker.worker.sync.apply.replay_stop_lsn = replorigin_session_origin_lsn;
-	namestrcpy(&worker.worker.sync.nspname, rv->schemaname);
-	namestrcpy(&worker.worker.sync.relname, rv->relname);
+	memcpy(&worker.worker.sync.nspname, nspname, sizeof(NameData));
+	memcpy(&worker.worker.sync.relname, relname, sizeof(NameData));
 
 	(void) pglogical_worker_register(&worker);
 }
