@@ -120,9 +120,35 @@ UserTableUpdateOpenIndexes(EState *estate, TupleTableSlot *slot)
 
 		/* FIXME: recheck the indexes */
 		if (recheckIndexes != NIL)
+		{
+			StringInfoData si;
+			ListCell *lc;
+			const char *idxname, *relname, *nspname;
+			Relation target_rel = estate->es_result_relation_info->ri_RelationDesc;
+
+			relname = RelationGetRelationName(target_rel);
+			nspname = get_namespace_name(RelationGetNamespace(target_rel));
+
+			initStringInfo(&si);
+			foreach (lc, recheckIndexes)
+			{
+				Oid idxoid = lfirst_oid(lc);
+				idxname = get_rel_name(idxoid);
+				if (idxname == NULL)
+					elog(ERROR, "cache lookup failed for index oid %u", idxoid);
+				if (si.len > 0)
+					appendStringInfoString(&si, ", ");
+				appendStringInfoString(&si, quote_identifier(idxname));
+			}
+
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					 errmsg("pglogical doesn't support index rechecks")));
+					 errmsg("pglogical doesn't support deferrable indexes"),
+					 errdetail("relation %s.%s has deferrable indexes: %s",
+								quote_identifier(nspname),
+								quote_identifier(relname),
+								si.data)));
+		}
 
 		list_free(recheckIndexes);
 	}
@@ -207,7 +233,7 @@ init_apply_exec_state(PGLogicalRelation *rel)
 	ApplyExecState	   *aestate = palloc0(sizeof(ApplyExecState));
 
 	/* Initialize the executor state. */
-	aestate->estate = create_estate_for_relation(rel->rel, rel->hasTriggers);
+	aestate->estate = create_estate_for_relation(rel->rel, true);
 	aestate->resultRelInfo = aestate->estate->es_result_relation_info;
 
 	aestate->slot = ExecInitExtraTupleSlot(aestate->estate);
@@ -247,13 +273,14 @@ void
 pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 {
 	ApplyExecState	   *aestate;
-	Oid					conflicts;
+	Oid					conflicts_idx_id;
 	TupleTableSlot	   *localslot;
 	HeapTuple			remotetuple;
 	HeapTuple			applytuple;
 	PGLogicalConflictResolution resolution;
 	List			   *recheckIndexes = NIL;
 	MemoryContext		oldctx;
+	bool				has_before_triggers = false;
 
 	/* Initialize the executor state. */
 	aestate = init_apply_exec_state(rel);
@@ -269,10 +296,14 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 #endif
 					);
 
-	/* Check for existing tuple with same key */
-	conflicts = pglogical_tuple_find_conflict(aestate->estate,
-											  newtup,
-											  localslot);
+	/*
+	 * Check for existing tuple with same key in any unique index containing
+	 * only normal columns. This doesn't just check the replica identity index,
+	 * but it'll prefer it and use it first.
+	 */
+	conflicts_idx_id = pglogical_tuple_find_conflict(aestate->estate,
+													 newtup,
+													 localslot);
 
 	/* Process and store remote tuple in the slot */
 	oldctx = MemoryContextSwitchTo(GetPerTupleMemoryContext(aestate->estate));
@@ -285,6 +316,8 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 	if (aestate->resultRelInfo->ri_TrigDesc &&
 		aestate->resultRelInfo->ri_TrigDesc->trig_insert_before_row)
 	{
+		has_before_triggers = true;
+
 		aestate->slot = ExecBRInsertTriggers(aestate->estate,
 											 aestate->resultRelInfo,
 											 aestate->slot);
@@ -301,16 +334,29 @@ pglogical_apply_heap_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 	/* trigger might have changed tuple */
 	remotetuple = ExecMaterializeSlot(aestate->slot);
 
-	if (OidIsValid(conflicts))
+	/* Did we find matching key in any candidate-key index? */
+	if (OidIsValid(conflicts_idx_id))
 	{
-		/* Tuple already exists, try resolving conflict. */
-		bool apply = try_resolve_conflict(rel->rel, localslot->tts_tuple,
-										  remotetuple, &applytuple,
-										  &resolution);
+		TransactionId		xmin;
+		TimestampTz			local_ts;
+		RepOriginId			local_origin;
+		bool				apply;
+		bool				local_origin_found;
 
-		pglogical_report_conflict(CONFLICT_INSERT_INSERT, rel->rel,
-								  localslot->tts_tuple, remotetuple,
-								  applytuple, resolution);
+		local_origin_found = get_tuple_origin(localslot->tts_tuple, &xmin,
+											  &local_origin, &local_ts);
+
+		/* Tuple already exists, try resolving conflict. */
+		apply = try_resolve_conflict(rel->rel, localslot->tts_tuple,
+									 remotetuple, &applytuple,
+									 &resolution);
+
+		pglogical_report_conflict(CONFLICT_INSERT_INSERT, rel,
+								  localslot->tts_tuple, NULL, remotetuple,
+								  applytuple, resolution, xmin,
+								  local_origin_found, local_origin,
+								  local_ts, conflicts_idx_id,
+								  has_before_triggers);
 
 		if (apply)
 		{
@@ -394,6 +440,8 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 	HeapTuple			remotetuple;
 	List			   *recheckIndexes = NIL;
 	MemoryContext		oldctx;
+	Oid					replident_idx_id;
+	bool				has_before_triggers = false;
 
 	/* Initialize the executor state. */
 	aestate = init_apply_exec_state(rel);
@@ -403,12 +451,14 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 	PushActiveSnapshot(GetTransactionSnapshot());
 
 	/* Search for existing tuple with same key */
-	found = pglogical_tuple_find_replidx(aestate->estate, oldtup, localslot);
+	found = pglogical_tuple_find_replidx(aestate->estate, oldtup, localslot,
+										 &replident_idx_id);
 
 	/*
-	 * Tuple found.
+	 * Tuple found, update the local tuple.
 	 *
-	 * Note this will fail if there are other conflicting unique indexes.
+	 * Note this will fail if there are other unique indexes and one or more of
+	 * them would be violated by the new tuple.
 	 */
 	if (found)
 	{
@@ -434,6 +484,8 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 		if (aestate->resultRelInfo->ri_TrigDesc &&
 			aestate->resultRelInfo->ri_TrigDesc->trig_update_before_row)
 		{
+			has_before_triggers = true;
+
 			aestate->slot = ExecBRUpdateTriggers(aestate->estate,
 												 &aestate->epqstate,
 												 aestate->resultRelInfo,
@@ -468,9 +520,12 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 										 remotetuple, &applytuple,
 										 &resolution);
 
-			pglogical_report_conflict(CONFLICT_UPDATE_UPDATE, rel->rel,
-									  localslot->tts_tuple, remotetuple,
-									  applytuple, resolution);
+			pglogical_report_conflict(CONFLICT_UPDATE_UPDATE, rel,
+									  localslot->tts_tuple, oldtup,
+									  remotetuple, applytuple, resolution,
+									  xmin, local_origin_found, local_origin,
+									  local_ts, replident_idx_id,
+									  has_before_triggers);
 
 			if (applytuple != remotetuple)
 				ExecStoreTuple(applytuple, aestate->slot, InvalidBuffer, false);
@@ -520,8 +575,11 @@ pglogical_apply_heap_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 		remotetuple = heap_form_tuple(RelationGetDescr(rel->rel),
 									  newtup->values,
 									  newtup->nulls);
-		pglogical_report_conflict(CONFLICT_UPDATE_DELETE, rel->rel, NULL,
-								  remotetuple, NULL, PGLogicalResolution_Skip);
+		pglogical_report_conflict(CONFLICT_UPDATE_DELETE, rel, NULL, oldtup,
+								  remotetuple, NULL, PGLogicalResolution_Skip,
+								  InvalidTransactionId, false,
+								  InvalidRepOriginId, (TimestampTz)0,
+								  replident_idx_id, has_before_triggers);
 	}
 
 	/* Cleanup. */
@@ -539,6 +597,8 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 {
 	ApplyExecState	   *aestate;
 	TupleTableSlot	   *localslot;
+	Oid					replident_idx_id;
+	bool				has_before_triggers = false;
 
 	/* Initialize the executor state. */
 	aestate = init_apply_exec_state(rel);
@@ -547,10 +607,11 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 
 	PushActiveSnapshot(GetTransactionSnapshot());
 
-	if (pglogical_tuple_find_replidx(aestate->estate, oldtup, localslot))
+	if (pglogical_tuple_find_replidx(aestate->estate, oldtup, localslot,
+									 &replident_idx_id))
 	{
 		if (aestate->resultRelInfo->ri_TrigDesc &&
-			aestate->resultRelInfo->ri_TrigDesc->trig_update_before_row)
+			aestate->resultRelInfo->ri_TrigDesc->trig_delete_before_row)
 		{
 			bool dodelete = ExecBRDeleteTriggers(aestate->estate,
 												 &aestate->epqstate,
@@ -558,6 +619,8 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 												 &localslot->tts_tuple->t_self,
 												 NULL);
 
+			has_before_triggers = true;
+
 			if (!dodelete)		/* "do nothing" */
 			{
 				PopActiveSnapshot();
@@ -579,8 +642,11 @@ pglogical_apply_heap_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 		/* The tuple to be deleted could not be found. */
 		HeapTuple remotetuple = heap_form_tuple(RelationGetDescr(rel->rel),
 												oldtup->values, oldtup->nulls);
-		pglogical_report_conflict(CONFLICT_DELETE_DELETE, rel->rel, NULL,
-								  remotetuple, NULL, PGLogicalResolution_Skip);
+		pglogical_report_conflict(CONFLICT_DELETE_DELETE, rel, NULL, oldtup,
+								  remotetuple, NULL, PGLogicalResolution_Skip,
+								  InvalidTransactionId, false,
+								  InvalidRepOriginId, (TimestampTz)0,
+								  replident_idx_id, has_before_triggers);
 	}
 
 	/* Cleanup. */
