@@ -38,6 +38,7 @@
 #include "utils/rel.h"
 #include "utils/builtins.h"
 
+#include "pglogical.h"
 #include "pglogical_apply_spi.h"
 #include "pglogical_conflict.h"
 
@@ -122,9 +123,11 @@ pglogical_apply_spi_insert(PGLogicalRelation *rel, PGLogicalTupleData *newtup)
 			continue;
 
 		if (narg > 0)
-			appendStringInfo(&cmd, ", %s", NameStr(desc->attrs[att]->attname));
+			appendStringInfo(&cmd, ", %s",
+					quote_identifier(NameStr(desc->attrs[att]->attname)));
 		else
-			appendStringInfo(&cmd, "%s", NameStr(desc->attrs[att]->attname));
+			appendStringInfo(&cmd, "%s",
+					quote_identifier(NameStr(desc->attrs[att]->attname)));
 		narg++;
 	}
 
@@ -192,10 +195,12 @@ pglogical_apply_spi_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 
 		if (narg > 0)
 			appendStringInfo(&cmd, ", %s = $%u",
-							 NameStr(desc->attrs[att]->attname), narg + 1);
+							 quote_identifier(NameStr(desc->attrs[att]->attname)),
+							 narg + 1);
 		else
 			appendStringInfo(&cmd, "%s = $%u",
-							 NameStr(desc->attrs[att]->attname), narg + 1);
+							 quote_identifier(NameStr(desc->attrs[att]->attname)),
+							 narg + 1);
 
 		argtypes[narg] = desc->attrs[att]->atttypid;
 		values[narg] = newtup->values[att];
@@ -214,10 +219,12 @@ pglogical_apply_spi_update(PGLogicalRelation *rel, PGLogicalTupleData *oldtup,
 
 		if (narg > firstarg)
 			appendStringInfo(&cmd, " AND %s = $%u",
-							 NameStr(desc->attrs[att]->attname), narg + 1);
+							 quote_identifier(NameStr(desc->attrs[att]->attname)),
+							 narg + 1);
 		else
 			appendStringInfo(&cmd, " %s = $%u",
-							 NameStr(desc->attrs[att]->attname), narg + 1);
+							 quote_identifier(NameStr(desc->attrs[att]->attname)),
+							 narg + 1);
 
 		argtypes[narg] = desc->attrs[att]->atttypid;
 		values[narg] = oldtup->values[att];
@@ -262,10 +269,12 @@ pglogical_apply_spi_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 
 		if (narg > 0)
 			appendStringInfo(&cmd, " AND %s = $%u",
-							 NameStr(desc->attrs[att]->attname), narg + 1);
+							 quote_identifier(NameStr(desc->attrs[att]->attname)),
+							 narg + 1);
 		else
 			appendStringInfo(&cmd, " %s = $%u",
-							 NameStr(desc->attrs[att]->attname), narg + 1);
+							 quote_identifier(NameStr(desc->attrs[att]->attname)),
+							 narg + 1);
 
 		argtypes[narg] = desc->attrs[att]->atttypid;
 		values[narg] = oldtup->values[att];
@@ -282,7 +291,7 @@ pglogical_apply_spi_delete(PGLogicalRelation *rel, PGLogicalTupleData *oldtup)
 
 
 /* We currently can't support multi insert using COPY on windows. */
-#ifndef WIN32
+#if !defined(WIN32) && !defined(PGL_NO_STDIN_ASSIGN)
 
 bool
 pglogical_apply_spi_can_mi(PGLogicalRelation *rel)
@@ -496,8 +505,8 @@ pglogical_proccess_copy(pglogical_copyState *pglcstate)
 	SPI_push();
 
 	/* Initiate the actual COPY */
-	DoCopy((CopyStmt *) linitial(pglcstate->copy_parsetree),
-			pglcstate->copy_stmt->data, &processed);
+	PGLDoCopy((CopyStmt *) linitial(pglcstate->copy_parsetree),
+			  pglcstate->copy_stmt->data, &processed);
 
 	/* Clean up SPI state */
 	SPI_pop();
