@@ -22,6 +22,8 @@
 #include "access/transam.h"
 #include "access/xact.h"
 
+#include "catalog/pg_type.h"
+
 #include "executor/executor.h"
 
 #include "parser/parse_relation.h"
@@ -31,11 +33,13 @@
 #include "storage/lmgr.h"
 
 #include "utils/builtins.h"
+#include "utils/datetime.h"
 #include "utils/lsyscache.h"
 #include "utils/rel.h"
 #include "utils/snapmgr.h"
 #include "utils/syscache.h"
 #include "utils/tqual.h"
+#include "utils/typcache.h"
 
 #include "pglogical_conflict.h"
 #include "pglogical_proto_native.h"
@@ -43,11 +47,15 @@
 int		pglogical_conflict_resolver = PGLOGICAL_RESOLVE_APPLY_REMOTE;
 int		pglogical_conflict_log_level = LOG;
 
+static void tuple_to_stringinfo(StringInfo s, TupleDesc tupdesc,
+	HeapTuple tuple);
+
 /*
  * Setup a ScanKey for a search in the relation 'rel' for a tuple 'key' that
  * is setup to match 'rel' (*NOT* idxrel!).
  *
- * Returns whether any column contains NULLs.
+ * Returns whether any column in the passed tuple contains a NULL for an
+ * indexed field.
  */
 static bool
 build_index_scan_key(ScanKey skey, Relation rel, Relation idxrel, PGLogicalTupleData *tup)
@@ -70,6 +78,10 @@ build_index_scan_key(ScanKey skey, Relation rel, Relation idxrel, PGLogicalTuple
 	Assert(!isnull);
 	indkey = (int2vector *) DatumGetPointer(indkeyDatum);
 
+	/*
+	 * Examine each indexed attribute to ensure the passed tuple's matching
+	 * value isn't NULL and we have an equality operator for it.
+	 */
 	for (attoff = 0; attoff < RelationGetNumberOfAttributes(idxrel); attoff++)
 	{
 		Oid			operator;
@@ -114,7 +126,7 @@ build_index_scan_key(ScanKey skey, Relation rel, Relation idxrel, PGLogicalTuple
  * Search the index 'idxrel' for a tuple identified by 'skey' in 'rel'.
  *
  * If a matching tuple is found lock it with lockmode, fill the slot with its
- * contents and return true, return false is returned otherwise.
+ * contents and return true, false is returned otherwise.
  */
 static bool
 find_index_tuple(ScanKey skey, Relation rel, Relation idxrel,
@@ -126,6 +138,11 @@ find_index_tuple(ScanKey skey, Relation rel, Relation idxrel,
 	SnapshotData snap;
 	TransactionId xwait;
 
+	/*
+	 * We need SnapshotDirty because we're doing uniqueness lookups that must
+	 * consider rows added/updated by concurrent transactions, just like a
+	 * normal UNIQUE check does.
+	 */
 	InitDirtySnapshot(snap);
 	scan = index_beginscan(rel, idxrel, &snap,
 						   RelationGetNumberOfAttributes(idxrel),
@@ -143,16 +160,22 @@ retry:
 		ExecStoreTuple(scantuple, slot, InvalidBuffer, false);
 		ExecMaterializeSlot(slot);
 
+		/*
+		 * Did any concurrent txn affect the tuple? (See
+		 * HeapTupleSatisfiesDirty for how we get this).
+		 */
 		xwait = TransactionIdIsValid(snap.xmin) ?
 			snap.xmin : snap.xmax;
 
 		if (TransactionIdIsValid(xwait))
 		{
+			/* Wait for the specified transaction to commit or abort */
 			XactLockTableWait(xwait, NULL, NULL, XLTW_None);
 			goto retry;
 		}
 	}
 
+	/* Matching tuple found, no concurrent txns modifying it */
 	if (found)
 	{
 		Buffer buf;
@@ -177,9 +200,16 @@ retry:
 		switch (res)
 		{
 			case HeapTupleMayBeUpdated:
+				/* lock was successfully acquired */
 				break;
 			case HeapTupleUpdated:
-				/* XXX: Improve handling here */
+				/*
+				 * We lost a race between when we looked up the tuple and
+				 * checked for concurrent modifying txns and when we tried to
+				 * lock the matched tuple.
+				 *
+				 * XXX: Improve handling here.
+				 */
 				ereport(LOG,
 						(errcode(ERRCODE_T_R_SERIALIZATION_FAILURE),
 						 errmsg("concurrent update, retrying")));
@@ -196,11 +226,14 @@ retry:
 }
 
 /*
- * Find tuple using REPLICA IDENTITY index.
+ * Find tuple using REPLICA IDENTITY index and output it in 'oldslot'
+ * if found.
+ *
+ * The index oid is also output.
  */
 bool
 pglogical_tuple_find_replidx(EState *estate, PGLogicalTupleData *tuple,
-							 TupleTableSlot *oldslot)
+							 TupleTableSlot *oldslot, Oid *idxrelid)
 {
 	ResultRelInfo  *relinfo = estate->es_result_relation_info;
 	Oid				idxoid;
@@ -212,15 +245,19 @@ pglogical_tuple_find_replidx(EState *estate, PGLogicalTupleData *tuple,
 	idxoid = RelationGetReplicaIndex(relinfo->ri_RelationDesc);
 	if (!OidIsValid(idxoid))
 	{
-		elog(ERROR, "could not find primary key for table with oid %u",
-			 RelationGetRelid(relinfo->ri_RelationDesc));
+		ereport(ERROR,
+				(errmsg("could not find REPLICA IDENTITY index for table %s with oid %u",
+						get_rel_name(RelationGetRelid(relinfo->ri_RelationDesc)),
+						RelationGetRelid(relinfo->ri_RelationDesc)),
+				 errhint("The REPLICA IDENTITY index is usually the PRIMARY KEY. See the PostgreSQL docs for ALTER TABLE ... REPLICA IDENTITY")));
 	}
+	*idxrelid = idxoid;
 	idxrel = index_open(idxoid, RowExclusiveLock);
 
-	/* Buold scan key for just opened index*/
+	/* Build scan key for just opened index*/
 	build_index_scan_key(index_key, relinfo->ri_RelationDesc, idxrel, tuple);
 
-	/* Try to find the row. */
+	/* Try to find the row and store any matching row in 'oldslot'. */
 	found = find_index_tuple(index_key, relinfo->ri_RelationDesc, idxrel,
 							 LockTupleExclusive, oldslot);
 
@@ -231,67 +268,112 @@ pglogical_tuple_find_replidx(EState *estate, PGLogicalTupleData *tuple,
 }
 
 /*
- * Find the tuple in a table using any index.
+ * Find the tuple in a table using any index and returns the conflicting
+ * index's oid, if any conflict found.
+ *
+ * This is not wholly safe. It does not consider the table's upstream replica
+ * identity, and may choose to resolve the conflict on a unique index that
+ * isn't part of the replica identity.
+ *
+ * Order-of-apply issues between multiple upstreams can lead to
+ * non-deterministic behaviour in cases where we resolve one conflict using one
+ * index, then a second conflict using a different index.
+ *
+ * We should really respect the replica identity more (i.e. use
+ * pglogical_tuple_find_replidx). Or at least raise a WARNING that an
+ * inconsistency may arise.
  */
 Oid
 pglogical_tuple_find_conflict(EState *estate, PGLogicalTupleData *tuple,
-							  TupleTableSlot *oldslot)
+							  TupleTableSlot *outslot)
 {
-	Oid		conflict_idx = InvalidOid;
-	ScanKeyData	index_key[INDEX_MAX_KEYS];
-	int			i;
-	ResultRelInfo *relinfo;
-	ItemPointerData conflicting_tid;
+	Oid				conflict_idx = InvalidOid;
+	ScanKeyData		index_key[INDEX_MAX_KEYS];
+	int				i;
+	ItemPointerData	conflicting_tid;
+	Oid				replidxoid;
+	bool			found = false;
+	ResultRelInfo  *relinfo = estate->es_result_relation_info;
 
 	ItemPointerSetInvalid(&conflicting_tid);
 
 	relinfo = estate->es_result_relation_info;
 
-	/* Do a SnapshotDirty search for conflicting tuples. */
+	/*
+	 * Check the replica identity index with a SnapshotDirty scan first, like
+	 * pglogical_tuple_find_replidx, but without ERRORing if we don't find
+	 * a replica identity index.
+	 */
+	replidxoid = RelationGetReplicaIndex(relinfo->ri_RelationDesc);
+	if (OidIsValid(replidxoid))
+	{
+		ScanKeyData	index_key[INDEX_MAX_KEYS];
+		Relation	idxrel = index_open(replidxoid, RowExclusiveLock);
+		build_index_scan_key(index_key, relinfo->ri_RelationDesc, idxrel, tuple);
+		found = find_index_tuple(index_key, relinfo->ri_RelationDesc, idxrel,
+							 LockTupleExclusive, outslot);
+		index_close(idxrel, NoLock);
+		if (found)
+			return replidxoid;
+	}
+
+	/*
+	 * Do a SnapshotDirty search for conflicting tuples. If any is found
+	 * store it in outslot and return the oid of the matching index. We
+	 * don't continue scanning for matches in other indexes, so we won't
+	 * notice if the tuple conflicts with another index, and it'll
+	 * raise a unique violation on apply instead.
+	 *
+	 * We could carry on here even if (found) and look for secondary conflicts,
+	 * but all we'd be able to do would be ERROR here instead of later. The
+	 * rest of the time we'd just pay a useless performance cost for extra
+	 * index scans.
+	 */
 	for (i = 0; i < relinfo->ri_NumIndices; i++)
 	{
 		IndexInfo  *ii = relinfo->ri_IndexRelationInfo[i];
 		Relation	idxrel;
-		bool found = false;
 
 		/*
 		 * Only unique indexes are of interest here, and we can't deal with
-		 * expression indexes so far. FIXME: predicates should be handled
-		 * better.
+		 * expression indexes so far.
 		 */
 		if (!ii->ii_Unique || ii->ii_Expressions != NIL)
 			continue;
 
+		/*
+		 * TODO: predicates should be handled better. There's no point scanning
+		 * an index where the predicates show it could never match anyway, and
+		 * it can produce false conflicts if the predicate includes non-indexed
+		 * columns. We could find a local tuple that matches the predicate in
+		 * the index, but there's only a true conflict if the remote tuple also
+		 * matches the predicate. If we ignore the predicate we generate a false
+		 * conflict. See RM#1839.
+		 *
+		 * For now we reject conflict resolution on indexes with predicates
+		 * entirely. If there's a conflict it'll be raised on apply with a
+		 * unique violation.
+		 */
+		if (ii->ii_Predicate != NIL)
+			continue;
+
 		idxrel = relinfo->ri_IndexRelationDescs[i];
 
+		/* No point re-scanning the replica identity index */
+		if (RelationGetRelid(idxrel) == replidxoid)
+			continue;
+
 		if (build_index_scan_key(index_key, relinfo->ri_RelationDesc,
 								 idxrel, tuple))
 			continue;
 
-		/* Try to find conflicting row. */
+		/* Try to find conflicting row and store in 'outslot' */
 		found = find_index_tuple(index_key, relinfo->ri_RelationDesc,
-								 idxrel, LockTupleExclusive, oldslot);
+								 idxrel, LockTupleExclusive, outslot);
 
-		/* Alert if there's more than one conflicting unique key, we can't
-		 * currently handle that situation. */
-		if (found &&
-			ItemPointerIsValid(&conflicting_tid) &&
-			!ItemPointerEquals(&oldslot->tts_tuple->t_self,
-							   &conflicting_tid))
+		if (found)
 		{
-			/* TODO: Report tuple identity in log */
-			ereport(ERROR,
-				(errcode(ERRCODE_UNIQUE_VIOLATION),
-				errmsg("multiple unique constraints violated by remote tuple"),
-				errdetail("cannot apply transaction because remotely tuple "
-						  "conflicts with a local tuple on more than one "
-						  "UNIQUE constraint and/or PRIMARY KEY"),
-				errhint("Resolve the conflict by removing or changing the "
-						"conflicting local tuple")));
-		}
-		else if (found)
-		{
-			ItemPointerCopy(&oldslot->tts_tuple->t_self, &conflicting_tid);
+			ItemPointerCopy(&outslot->tts_tuple->t_self, &conflicting_tid);
 			conflict_idx = RelationGetRelid(idxrel);
 			break;
 		}
@@ -371,16 +453,15 @@ get_tuple_origin(HeapTuple local_tuple, TransactionId *xmin,
 				 RepOriginId *local_origin, TimestampTz *local_ts)
 {
 
+	*xmin = HeapTupleHeaderGetXmin(local_tuple->t_data);
 	if (!track_commit_timestamp)
 	{
-		*xmin = HeapTupleHeaderGetXmin(local_tuple->t_data);
 		*local_origin = replorigin_session_origin;
 		*local_ts = replorigin_session_origin_timestamp;
 		return false;
 	}
 	else
 	{
-		*xmin = HeapTupleHeaderGetXmin(local_tuple->t_data);
 		if (TransactionIdIsValid(*xmin) && !TransactionIdIsNormal(*xmin))
 		{
 			/*
@@ -498,35 +579,112 @@ conflict_resolution_to_string(PGLogicalConflictResolution resolution)
 /*
  * Log the conflict to server log.
  *
- * TODO: provide more detail.
+ * There are number of tuples passed:
+ *
+ * - The local tuple we conflict with or NULL if not found [localtuple];
+ *
+ * - If the remote tuple was an update, the key of the old tuple
+ *   as a PGLogicalTuple [oldkey]
+ *
+ * - The remote tuple, after we fill any defaults and apply any local
+ *   BEFORE triggers but before conflict resolution [remotetuple];
+ *
+ * - The tuple we'll actually apply if any, after conflict resolution
+ *   [applytuple]
+ *
+ * The PGLogicalRelation's name info is for the remote rel. If we add relation
+ * mapping we'll need to get the name/namespace of the local relation too.
+ *
+ * This runs in MessageContext so we don't have to worry about leaks, but
+ * we still try to free the big chunks as we go.
  */
 void
-pglogical_report_conflict(PGLogicalConflictType conflict_type, Relation rel,
-						  HeapTuple localtuple, HeapTuple remotetuple,
+pglogical_report_conflict(PGLogicalConflictType conflict_type,
+						  PGLogicalRelation *rel,
+						  HeapTuple localtuple,
+						  PGLogicalTupleData *oldkey,
+						  HeapTuple remotetuple,
 						  HeapTuple applytuple,
-						  PGLogicalConflictResolution resolution)
+						  PGLogicalConflictResolution resolution,
+						  TransactionId local_tuple_xid,
+						  bool found_local_origin,
+						  RepOriginId local_tuple_origin,
+						  TimestampTz local_tuple_commit_ts,
+						  Oid conflict_idx_oid,
+						  bool has_before_triggers)
 {
+	char local_tup_ts_str[MAXDATELEN] = "(unset)";
+	StringInfoData localtup, remotetup;
+	TupleDesc desc = RelationGetDescr(rel->rel);
+	const char *idxname = "(unknown)";
+	const char *qualrelname;
+
+	memset(local_tup_ts_str, 0, MAXDATELEN);
+	if (found_local_origin)
+		strcpy(local_tup_ts_str,
+			timestamptz_to_str(local_tuple_commit_ts));
+
+	initStringInfo(&remotetup);
+	tuple_to_stringinfo(&remotetup, desc, remotetuple);
+
+	if (localtuple != NULL)
+	{
+		initStringInfo(&localtup);
+		tuple_to_stringinfo(&localtup, desc, localtuple);
+	}
+
+	if (OidIsValid(conflict_idx_oid))
+		idxname = get_rel_name(conflict_idx_oid);
+
+	qualrelname = quote_qualified_identifier(
+		get_namespace_name(RelationGetNamespace(rel->rel)),
+		RelationGetRelationName(rel->rel));
+
+	/*
+	 * We try to provide a lot of information about conflicting tuples because
+	 * the conflicts are often transient and timing-sensitive. It's rare that
+	 * we can examine a stopped system or reproduce them at leisure. So the
+	 * more info we have in the logs, the better chance we have of diagnosing
+	 * application issues. It's worth paying the price of some log spam.
+	 *
+	 * This deliberately somewhat overlaps with the context info we log with
+	 * log_error_verbosity=verbose because we don't necessarily have all that
+	 * info enabled.
+	 */
 	switch (conflict_type)
 	{
 		case CONFLICT_INSERT_INSERT:
 		case CONFLICT_UPDATE_UPDATE:
 			ereport(pglogical_conflict_log_level,
 					(errcode(ERRCODE_INTEGRITY_CONSTRAINT_VIOLATION),
-					 errmsg("CONFLICT: remote %s on relation %s. Resolution: %s.",
-							CONFLICT_INSERT_INSERT ? "INSERT" : "UPDATE",
-							quote_qualified_identifier(get_namespace_name(RelationGetNamespace(rel)),
-													   RelationGetRelationName(rel)),
-							conflict_resolution_to_string(resolution))));
+					 errmsg("CONFLICT: remote %s on relation %s (local index %s). Resolution: %s.",
+							conflict_type == CONFLICT_INSERT_INSERT ? "INSERT" : "UPDATE",
+							qualrelname, idxname,
+							conflict_resolution_to_string(resolution)),
+					 errdetail("existing local tuple {%s} xid=%u,origin=%d,timestamp=%s; remote tuple {%s}%s in xact origin=%u,timestamp=%s,commit_lsn=%X/%X",
+							   localtup.data, local_tuple_xid,
+							   found_local_origin ? (int)local_tuple_origin : -1,
+							   local_tup_ts_str,
+							   remotetup.data, has_before_triggers ? "*":"",
+							   replorigin_session_origin,
+							   timestamptz_to_str(replorigin_session_origin_timestamp),
+							   (uint32)(replorigin_session_origin_lsn<<32),
+							   (uint32)replorigin_session_origin_lsn)));
 			break;
 		case CONFLICT_UPDATE_DELETE:
 		case CONFLICT_DELETE_DELETE:
 			ereport(pglogical_conflict_log_level,
 					(errcode(ERRCODE_INTEGRITY_CONSTRAINT_VIOLATION),
-					 errmsg("CONFLICT: remote %s on relation %s (tuple not found). Resolution: %s.",
-							CONFLICT_UPDATE_DELETE ? "UPDATE" : "DELETE",
-							quote_qualified_identifier(get_namespace_name(RelationGetNamespace(rel)),
-													   RelationGetRelationName(rel)),
-							conflict_resolution_to_string(resolution))));
+					 errmsg("CONFLICT: remote %s on relation %s replica identity index %s (tuple not found). Resolution: %s.",
+							conflict_type == CONFLICT_UPDATE_DELETE ? "UPDATE" : "DELETE",
+							qualrelname, idxname,
+							conflict_resolution_to_string(resolution)),
+					 errdetail("remote tuple {%s}%s in xact origin=%u,timestamp=%s,commit_lsn=%X/%X",
+							   remotetup.data, has_before_triggers ? "*":"",
+							   replorigin_session_origin,
+							   timestamptz_to_str(replorigin_session_origin_timestamp),
+							   (uint32)(replorigin_session_origin_lsn<<32),
+							   (uint32)replorigin_session_origin_lsn)));
 			break;
 	}
 }
@@ -551,3 +709,111 @@ pglogical_conflict_resolver_check_hook(int *newval, void **extra,
 
 	return true;
 }
+
+
+/*
+ * print the tuple 'tuple' into the StringInfo s
+ *
+ * (Based on bdr2)
+ */
+static void
+tuple_to_stringinfo(StringInfo s, TupleDesc tupdesc, HeapTuple tuple)
+{
+	int			natt;
+	Oid			oid;
+	bool		first = true;
+
+	static const int MAX_CONFLICT_LOG_ATTR_LEN = 40;
+
+	/* print oid of tuple, it's not included in the TupleDesc */
+	if ((oid = HeapTupleHeaderGetOid(tuple->t_data)) != InvalidOid)
+	{
+		appendStringInfo(s, "oid[oid]:%u", oid);
+	}
+
+	/* print all columns individually */
+	for (natt = 0; natt < tupdesc->natts; natt++)
+	{
+		Form_pg_attribute attr; /* the attribute itself */
+		Oid			typid;		/* type of current attribute */
+		HeapTuple	type_tuple; /* information about a type */
+		Form_pg_type type_form;
+		Oid			typoutput;	/* output function */
+		bool		typisvarlena;
+		Datum		origval;	/* possibly toasted Datum */
+		Datum		val	= PointerGetDatum(NULL); /* definitely detoasted Datum */
+		char	   *outputstr = NULL;
+		bool		isnull;		/* column is null? */
+
+		attr = tupdesc->attrs[natt];
+
+		/*
+		 * don't print dropped columns, we can't be sure everything is
+		 * available for them
+		 */
+		if (attr->attisdropped)
+			continue;
+
+		/*
+		 * Don't print system columns
+		 */
+		if (attr->attnum < 0)
+			continue;
+
+		typid = attr->atttypid;
+
+		/* gather type name */
+		type_tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(typid));
+		if (!HeapTupleIsValid(type_tuple))
+			elog(ERROR, "cache lookup failed for type %u", typid);
+		type_form = (Form_pg_type) GETSTRUCT(type_tuple);
+
+		/* print attribute name */
+		if (first)
+			first = false;
+		else
+			appendStringInfoChar(s, ' ');
+		appendStringInfoString(s, NameStr(attr->attname));
+
+		/* print attribute type */
+		appendStringInfoChar(s, '[');
+		appendStringInfoString(s, NameStr(type_form->typname));
+		appendStringInfoChar(s, ']');
+
+		/* query output function */
+		getTypeOutputInfo(typid,
+						  &typoutput, &typisvarlena);
+
+		ReleaseSysCache(type_tuple);
+
+		/* get Datum from tuple */
+		origval = heap_getattr(tuple, natt + 1, tupdesc, &isnull);
+
+		if (isnull)
+			outputstr = "(null)";
+		else if (typisvarlena && VARATT_IS_EXTERNAL_ONDISK(origval))
+			outputstr = "(unchanged-toast-datum)";
+		else if (typisvarlena)
+			val = PointerGetDatum(PG_DETOAST_DATUM(origval));
+		else
+			val = origval;
+
+		/* print data */
+		if (outputstr == NULL)
+			outputstr = OidOutputFunctionCall(typoutput, val);
+
+		/*
+		 * Abbreviate the Datum if it's too long. This may make it syntatically
+		 * invalid, but it's not like we're writing out a valid ROW(...) as it
+		 * is.
+		 */
+		if (strlen(outputstr) > MAX_CONFLICT_LOG_ATTR_LEN)
+		{
+			/* The null written at the end of strcpy will truncate the string */
+			strcpy(&outputstr[MAX_CONFLICT_LOG_ATTR_LEN-5], "...");
+		}
+
+		appendStringInfoChar(s, ':');
+		appendStringInfoString(s, outputstr);
+	}
+}
