@@ -506,6 +506,8 @@ main(int argc, char **argv)
 		/* Create the extension. */
 		print_msg(VERBOSITY_VERBOSE,
 				  _("Creating pglogical extension for database %s...\n"), db);
+		if (PQserverVersion(subscriber_conn) < 90500)
+			install_extension(subscriber_conn, "pglogical_origin");
 		install_extension(subscriber_conn, "pglogical");
 
 		/*
@@ -844,7 +846,7 @@ get_remote_info(PGconn* conn)
 
 	res = PQexec(conn, "SELECT node_id, node_name, sysid, dbname, replication_sets FROM pglogical.pglogical_node_info()");
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
-		die(_("could fetch remote node info: %s\n"), PQerrorMessage(conn));
+		die(_("could not fetch remote node info: %s\n"), PQerrorMessage(conn));
 
 	/* No nodes found? */
 	if (PQntuples(res) == 0)
@@ -926,14 +928,20 @@ remove_unwanted_data(PGconn *conn)
 {
 	PGresult		   *res;
 
-	/* Remove replication identifiers. */
-	res = PQexec(conn, "SELECT pg_replication_origin_drop(external_id) FROM pg_replication_origin_status;");
-	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+	/*
+	 * Remove replication identifiers (9.4 will get the removed by dropping
+	 * the extension later as we emulate them there).
+	 */
+	if (PQserverVersion(conn) >= 90500)
 	{
+		res = PQexec(conn, "SELECT pg_replication_origin_drop(external_id) FROM pg_replication_origin_status;");
+		if (PQresultStatus(res) != PGRES_TUPLES_OK)
+		{
+			PQclear(res);
+			die(_("Could not remove existing replication origins: %s\n"), PQerrorMessage(conn));
+		}
 		PQclear(res);
-		die(_("Could not remove existing replication origins: %s\n"), PQerrorMessage(conn));
 	}
-	PQclear(res);
 
 	res = PQexec(conn, "DROP EXTENSION pglogical CASCADE;");
 	if (PQresultStatus(res) != PGRES_COMMAND_OK)
@@ -953,32 +961,51 @@ initialize_replication_origin(PGconn *conn, char *origin_name, char *remote_lsn)
 	PGresult   *res;
 	PQExpBuffer query = createPQExpBuffer();
 
-	printfPQExpBuffer(query, "SELECT pg_replication_origin_create(%s)",
-					  PQescapeLiteral(conn, origin_name, strlen(origin_name)));
-
-	res = PQexec(conn, query->data);
-
-	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+	if (PQserverVersion(conn) >= 90500)
 	{
-		die(_("Could not create replication origin \"%s\": status %s: %s\n"),
-			 query->data,
-			 PQresStatus(PQresultStatus(res)), PQresultErrorMessage(res));
-	}
-	PQclear(res);
-
-	if (remote_lsn)
-	{
-		printfPQExpBuffer(query, "SELECT pg_replication_origin_advance(%s, '%s')",
-						 PQescapeLiteral(conn, origin_name, strlen(origin_name)),
-						 remote_lsn);
+		printfPQExpBuffer(query, "SELECT pg_replication_origin_create(%s)",
+						  PQescapeLiteral(conn, origin_name, strlen(origin_name)));
 
 		res = PQexec(conn, query->data);
 
 		if (PQresultStatus(res) != PGRES_TUPLES_OK)
 		{
-			die(_("Could not advance replication origin \"%s\": status %s: %s\n"),
-				 query->data,
-				 PQresStatus(PQresultStatus(res)), PQresultErrorMessage(res));
+			die(_("Could not create replication origin \"%s\": status %s: %s\n"),
+				query->data,
+				PQresStatus(PQresultStatus(res)), PQresultErrorMessage(res));
+		}
+		PQclear(res);
+
+		if (remote_lsn)
+		{
+			printfPQExpBuffer(query, "SELECT pg_replication_origin_advance(%s, '%s')",
+							  PQescapeLiteral(conn, origin_name, strlen(origin_name)),
+							  remote_lsn);
+
+			res = PQexec(conn, query->data);
+
+			if (PQresultStatus(res) != PGRES_TUPLES_OK)
+			{
+				die(_("Could not advance replication origin \"%s\": status %s: %s\n"),
+					query->data,
+					PQresStatus(PQresultStatus(res)), PQresultErrorMessage(res));
+			}
+			PQclear(res);
+		}
+	}
+	else
+	{
+		printfPQExpBuffer(query, "INSERT INTO pglogical_origin.replication_origin (roident, roname, roremote_lsn) SELECT COALESCE(MAX(roident::int), 0) + 1, %s, %s FROM pglogical_origin.replication_origin",
+						  PQescapeLiteral(conn, origin_name, strlen(origin_name)),
+						  remote_lsn ? PQescapeLiteral(conn, remote_lsn, strlen(remote_lsn)) : "0");
+
+		res = PQexec(conn, query->data);
+
+		if (PQresultStatus(res) != PGRES_COMMAND_OK)
+		{
+			die(_("Could not create replication origin \"%s\": status %s: %s\n"),
+				query->data,
+				PQresStatus(PQresultStatus(res)), PQresultErrorMessage(res));
 		}
 		PQclear(res);
 	}
