@@ -25,8 +25,8 @@ find_other_exec_version(const char *argv0, const char *target,
 	char		cmd[MAXPGPATH];
 	char		cmd_output[1024];
 	FILE       *output;
-	int			pre_dot,
-				post_dot;
+	int			pre_dot = 0,
+				post_dot = 0;
 
 	if (find_my_exec(argv0, retpath) < 0)
 		return -1;
@@ -39,7 +39,8 @@ find_other_exec_version(const char *argv0, const char *target,
 	snprintf(retpath + strlen(retpath), MAXPGPATH - strlen(retpath),
 			 "/%s%s", target, EXE);
 
-	snprintf(cmd, sizeof(cmd), "\"%s\" -V", retpath);
+	/* And request the version from the program */
+	snprintf(cmd, sizeof(cmd), "\"%s\" --version", retpath);
 
 	if ((output = popen(cmd, "r")) == NULL)
 		return -1;
@@ -51,10 +52,17 @@ find_other_exec_version(const char *argv0, const char *target,
 	}
 	pclose(output);
 
-	if (sscanf(cmd_output, "%*s %*s %d.%d", &pre_dot, &post_dot) != 2)
+	if (sscanf(cmd_output, "%*s %*s %d.%d", &pre_dot, &post_dot) < 1)
 		return -2;
 
-	*version = (pre_dot * 100 + post_dot) * 100;
+	/*
+	  similar to version number exposed by "server_version_num" but without
+	  the minor :
+	  9.6.1 -> 90601  -> 90600
+	  10.1  -> 100001 -> 100000)
+	*/
+	*version = (pre_dot < 10) ?
+	  (pre_dot * 100 + post_dot) * 100 : pre_dot * 100 * 100;
 
 	return 0;
 }
