@@ -11,6 +11,12 @@
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
+
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <dirent.h>
+
 #include "pglogical_output_plugin.h"
 
 #include "mb/pg_wchar.h"
@@ -22,6 +28,7 @@
 #include "access/xact.h"
 #include "executor/executor.h"
 #include "catalog/namespace.h"
+#include "storage/fd.h"
 #include "utils/inval.h"
 #include "utils/memutils.h"
 #include "utils/rel.h"
@@ -75,6 +82,7 @@ static PGLRelMetaCacheEntry *relmetacache_get_relation(PGLogicalOutputData *data
 static void relmetacache_destroy(void);
 static void relmetacache_prune(void);
 
+static void pglReorderBufferCleanSerializedTXNs(const char *slotname);
 
 /* specify output plugin callbacks */
 void
@@ -180,6 +188,17 @@ pg_decode_startup(LogicalDecodingContext * ctx, OutputPluginOptions *opt,
 		PGLogicalLocalNode *node;
 		MemoryContext oldctx;
 
+		/*
+		 * There's a potential corruption bug in PostgreSQL 10.1, 9.6.6, 9.5.10
+		 * and 9.4.15 that can cause reorder buffers to accumulate duplicated
+		 * transactions. See
+		 *   https://www.postgresql.org/message-id/CAMsr+YHdX=XECbZshDZ2CZNWGTyw-taYBnzqVfx4JzM4ExP5xg@mail.gmail.com
+		 *
+		 * We can defend against this by doing our own cleanup of any serialized
+		 * txns in the reorder buffer on startup.
+		 */
+		pglReorderBufferCleanSerializedTXNs(NameStr(MyReplicationSlot->data.name));
+
 		if (!IsTransactionState())
 		{
 			StartTransactionCommand();
@@ -925,3 +944,44 @@ relmetacache_prune(void)
 
 	InvalidRelMetaCacheCnt = 0;
 }
+
+/*
+ * Clone of ReorderBufferCleanSerializedTXNs; see
+ * https://www.postgresql.org/message-id/CAMsr+YHdX=XECbZshDZ2CZNWGTyw-taYBnzqVfx4JzM4ExP5xg@mail.gmail.com
+ */
+static void
+pglReorderBufferCleanSerializedTXNs(const char *slotname)
+{
+	DIR		   *spill_dir = NULL;
+	struct dirent *spill_de;
+	struct stat statbuf;
+	char		path[MAXPGPATH * 2 + 12];
+
+	sprintf(path, "pg_replslot/%s", slotname);
+
+	/* we're only handling directories here, skip if it's not our's */
+	if (lstat(path, &statbuf) == 0 && !S_ISDIR(statbuf.st_mode))
+		return;
+
+	spill_dir = AllocateDir(path);
+	while ((spill_de = ReadDir(spill_dir, path)) != NULL)
+	{
+		if (strcmp(spill_de->d_name, ".") == 0 ||
+			strcmp(spill_de->d_name, "..") == 0)
+			continue;
+
+		/* only look at names that can be ours */
+		if (strncmp(spill_de->d_name, "xid", 3) == 0)
+		{
+			sprintf(path, "pg_replslot/%s/%s", slotname,
+					spill_de->d_name);
+
+			if (unlink(path) != 0)
+				ereport(PANIC,
+						(errcode_for_file_access(),
+						 errmsg("could not remove file \"%s\": %m",
+								path)));
+		}
+	}
+	FreeDir(spill_dir);
+}
