@@ -39,6 +39,8 @@
 #include "nodes/makefuncs.h"
 #include "nodes/parsenodes.h"
 
+#include "pgstat.h"
+
 #include "replication/origin.h"
 
 #include "storage/fd.h"
@@ -70,13 +72,13 @@
 #endif
 #define PGRESTORE_BINARY "pg_restore"
 
-#define Natts_local_sync_state	5
+#define Natts_local_sync_state	6
 #define Anum_sync_kind			1
 #define Anum_sync_subid			2
 #define Anum_sync_nspname		3
 #define Anum_sync_relname		4
 #define Anum_sync_status		5
-
+#define Anum_sync_statuslsn		6
 
 void pglogical_sync_main(Datum main_arg);
 
@@ -93,7 +95,7 @@ dump_structure(PGLogicalSubscription *sub, const char *destfile,
 	StringInfoData	schema_filter;
 	StringInfoData	command;
 
-	if (find_other_exec_version(my_exec_path, PGDUMP_BINARY, &version, pg_dump))
+	if (find_other_exec_version(my_exec_path, PGDUMP_BINARY, &version, pg_dump) != 0)
 		elog(ERROR, "pglogical subscriber init failed to find pg_dump relative to binary %s",
 			 my_exec_path);
 
@@ -133,7 +135,7 @@ restore_structure(PGLogicalSubscription *sub, const char *srcfile,
 	int			res;
 	StringInfoData	command;
 
-	if (find_other_exec_version(my_exec_path, PGRESTORE_BINARY, &version, pg_restore))
+	if (find_other_exec_version(my_exec_path, PGRESTORE_BINARY, &version, pg_restore) != 0)
 		elog(ERROR, "pglogical subscriber init failed to find pg_restore relative to binary %s",
 			 my_exec_path);
 
@@ -319,6 +321,20 @@ finish_copy_target_tx(PGconn *conn)
 		elog(ERROR, "COMMIT on target node failed: %s",
 				PQresultErrorMessage(res));
 	PQclear(res);
+
+	/*
+	 * Resetting the origin explicitly before the backend exits will help
+	 * prevent races with other accesses to the same replication origin.
+	 */
+	if (PQserverVersion(conn) >= 90500)
+	{
+		res = PQexec(conn, "SELECT pg_catalog.pg_replication_origin_session_reset();\n");
+		if (PQresultStatus(res) != PGRES_TUPLES_OK)
+			elog(WARNING, "Resetting session origin on target node failed: %s",
+					PQresultErrorMessage(res));
+		PQclear(res);
+	}
+
 	PQfinish(conn);
 }
 
@@ -398,7 +414,9 @@ copy_table_data(PGconn *origin_conn, PGconn *target_conn,
 			first = false;
 		else
 			appendStringInfoString(&attlist, ",");
-		appendStringInfoString(&attlist, attname);
+		appendStringInfoString(&attlist,
+							   PQescapeIdentifier(origin_conn, attname,
+												  strlen(attname)));
 	}
 	MemoryContextSwitchTo(oldctx);
 	pglogical_relation_close(rel, AccessShareLock);
@@ -728,10 +746,10 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 
 		origin_conn = pglogical_connect(sub->origin_if->dsn,
 										sub->name, "snap");
-		use_failover_slot = pglogical_remote_function_exists(origin_conn,
-															 "pg_catalog",
-										  "pg_create_logical_replication_slot",
-															 3);
+		use_failover_slot = PQserverVersion(origin_conn) < 100000 &&
+			pglogical_remote_function_exists(origin_conn, "pg_catalog",
+											 "pg_create_logical_replication_slot",
+											 3);
 		PQfinish(origin_conn);
 
 		origin_conn_repl = pglogical_connect_replica(sub->origin_if->dsn,
@@ -753,14 +771,24 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 			PG_ENSURE_ERROR_CLEANUP(pglogical_sync_tmpfile_cleanup_cb,
 									CStringGetDatum(tmpfile));
 			{
+#if PG_VERSION_NUM >= 90500
+				Relation replorigin_rel;
+#endif
+
 				StartTransactionCommand();
 
 				originid = ensure_replication_origin(sub->slot_name);
 				elog(DEBUG3, "advancing origin with oid %u for forwarded row to %X/%X during subscription sync",
 					originid,
 					(uint32)(XactLastCommitEnd>>32), (uint32)XactLastCommitEnd);
+#if PG_VERSION_NUM >= 90500
+				replorigin_rel = heap_open(ReplicationOriginRelationId, RowExclusiveLock);
+#endif
 				replorigin_advance(originid, lsn, XactLastCommitEnd, true,
 								   true);
+#if PG_VERSION_NUM >= 90500
+				heap_close(replorigin_rel, RowExclusiveLock);
+#endif
 
 				CommitTransactionCommand();
 
@@ -814,7 +842,8 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 						{
 							set_table_sync_status(sub->id, remoterel->nspname,
 												  remoterel->relname,
-												  SYNC_STATUS_READY);
+												  SYNC_STATUS_READY,
+												  lsn);
 						}
 						else
 						{
@@ -822,9 +851,10 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 
 							newsync.kind = SYNC_KIND_FULL;
 							newsync.subid = sub->id;
-							newsync.nspname = remoterel->nspname;
-							newsync.relname = remoterel->relname;
+							namestrcpy(&newsync.nspname, remoterel->nspname);
+							namestrcpy(&newsync.relname, remoterel->relname);
 							newsync.status = SYNC_STATUS_READY;
+							newsync.statuslsn = lsn;
 							create_local_sync_status(&newsync);
 						}
 					}
@@ -875,9 +905,9 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 }
 
 char
-pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table)
+pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table,
+					 XLogRecPtr *status_lsn)
 {
-	XLogRecPtr	lsn;
 	PGconn	   *origin_conn_repl;
 	RepOriginId	originid;
 	char	   *snapshot;
@@ -895,14 +925,17 @@ pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table)
 
 	/* Check current state of the table. */
 	sync = get_table_sync_status(sub->id, table->schemaname, table->relname, false);
+	*status_lsn = sync->statuslsn;
 
 	/* Already synchronized, nothing to do here. */
-	if (sync->status == SYNC_STATUS_READY)
-		return SYNC_STATUS_READY;
+	if (sync->status == SYNC_STATUS_READY ||
+		sync->status == SYNC_STATUS_SYNCDONE)
+		return sync->status;
 
 	/* If previous sync attempt failed, we need to start from beginning. */
 	if (sync->status != SYNC_STATUS_INIT)
-		set_table_sync_status(sub->id, table->schemaname, table->relname, SYNC_STATUS_INIT);
+		set_table_sync_status(sub->id, table->schemaname, table->relname,
+							  SYNC_STATUS_INIT, InvalidXLogRecPtr);
 
 	CommitTransactionCommand();
 
@@ -910,21 +943,34 @@ pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table)
 												 sub->name, "copy");
 
 	snapshot = ensure_replication_slot_snapshot(origin_conn_repl,
-												sub->slot_name, false, &lsn);
+												sub->slot_name, false,
+												status_lsn);
 
 	/* Make sure we cleanup the slot if something goes wrong. */
 	PG_ENSURE_ERROR_CLEANUP(pglogical_sync_worker_cleanup_error_cb,
 							PointerGetDatum(sub));
 	{
+#if PG_VERSION_NUM >= 90500
+		Relation replorigin_rel;
+#endif
+
 		StartTransactionCommand();
 		originid = ensure_replication_origin(sub->slot_name);
 		elog(DEBUG2, "advancing origin %s (oid %u) for forwarded row to %X/%X after sync error",
 			MySubscription->slot_name, originid,
 			(uint32)(XactLastCommitEnd>>32), (uint32)XactLastCommitEnd);
-		replorigin_advance(originid, lsn, XactLastCommitEnd, true, true);
+
+#if PG_VERSION_NUM >= 90500
+		replorigin_rel = heap_open(ReplicationOriginRelationId, RowExclusiveLock);
+#endif
+		replorigin_advance(originid, *status_lsn, XactLastCommitEnd, true,
+						   true);
+#if PG_VERSION_NUM >= 90500
+		heap_close(replorigin_rel, RowExclusiveLock);
+#endif
 
 		set_table_sync_status(sub->id, table->schemaname, table->relname,
-							  SYNC_STATUS_DATA);
+							  SYNC_STATUS_DATA, *status_lsn);
 		CommitTransactionCommand();
 
 		/* Copy data. */
@@ -945,13 +991,20 @@ pglogical_sync_worker_finish(void)
 {
 	PGLogicalWorker	   *apply;
 
-	StartTransactionCommand();
-	/* Mark local table as ready. */
-	set_table_sync_status(MyApplyWorker->subid,
-						  NameStr(MyPGLogicalWorker->worker.sync.nspname),
-						  NameStr(MyPGLogicalWorker->worker.sync.relname),
-						  SYNC_STATUS_READY);
+	/*
+	 * Commit any outstanding transaction. This is the usual case, unless
+	 * there was nothing to do for the table.
+	 */
+	if (IsTransactionState())
+	{
+		CommitTransactionCommand();
+		pgstat_report_stat(false);
+	}
 
+	/* And flush all writes. */
+	XLogFlush(GetXLogWriteRecPtr());
+
+	StartTransactionCommand();
 	pglogical_sync_worker_cleanup(MySubscription);
 	CommitTransactionCommand();
 
@@ -977,11 +1030,13 @@ pglogical_sync_main(Datum main_arg)
 	int				slot = DatumGetInt32(main_arg);
 	PGconn		   *streamConn;
 	RepOriginId		originid;
-	XLogRecPtr		origin_startpos;
+	XLogRecPtr		lsn;
+	XLogRecPtr		status_lsn;
 	StringInfoData	slot_name;
 	RangeVar	   *copytable = NULL;
 	MemoryContext	saved_ctx;
 	char		   *tablename;
+	char			status;
 
 	/* Setup shmem. */
 	pglogical_worker_attach(slot, PGLOGICAL_WORKER_SYNC);
@@ -1041,7 +1096,8 @@ pglogical_sync_main(Datum main_arg)
 		 MySubscription->origin_if->name, MySubscription->origin_if->dsn);
 
 	/* Do the initial sync first. */
-	if (pglogical_sync_table(MySubscription, copytable) == SYNC_STATUS_READY)
+	status = pglogical_sync_table(MySubscription, copytable, &status_lsn);
+	if (status == SYNC_STATUS_SYNCDONE || status == SYNC_STATUS_READY)
 	{
 		pglogical_sync_worker_finish();
 		proc_exit(0);
@@ -1050,11 +1106,14 @@ pglogical_sync_main(Datum main_arg)
 	/* Wait for ack from the main apply thread. */
 	StartTransactionCommand();
 	set_table_sync_status(MySubscription->id, copytable->schemaname,
-						  copytable->relname, SYNC_STATUS_SYNCWAIT);
+						  copytable->relname, SYNC_STATUS_SYNCWAIT,
+						  status_lsn);
 	CommitTransactionCommand();
 
 	wait_for_sync_status_change(MySubscription->id, copytable->schemaname,
-								copytable->relname, SYNC_STATUS_CATCHUP);
+								copytable->relname, SYNC_STATUS_CATCHUP,
+								&lsn);
+	Assert(lsn == status_lsn);
 
 	/* Setup the origin and get the starting position for the replication. */
 	StartTransactionCommand();
@@ -1063,16 +1122,25 @@ pglogical_sync_main(Datum main_arg)
 		MySubscription->slot_name, originid);
 	replorigin_session_setup(originid);
 	replorigin_session_origin = originid;
-	origin_startpos = replorigin_session_get_progress(false);
-	CommitTransactionCommand();
+	Assert(status_lsn == replorigin_session_get_progress(false));
 
-	/* In case there is nothing to catchup, finish immediately. */
-	if (origin_startpos >= MyApplyWorker->replay_stop_lsn)
+	/*
+	 * In case there is nothing to catchup, finish immediately.
+	 * Note pglogical_sync_worker_finish() will commit.
+	 */
+	if (status_lsn >= MyApplyWorker->replay_stop_lsn)
 	{
+		/* Mark local table as done. */
+		set_table_sync_status(MyApplyWorker->subid,
+							  NameStr(MyPGLogicalWorker->worker.sync.nspname),
+							  NameStr(MyPGLogicalWorker->worker.sync.relname),
+							  SYNC_STATUS_SYNCDONE, status_lsn);
 		pglogical_sync_worker_finish();
 		proc_exit(0);
 	}
 
+	CommitTransactionCommand();
+
 	/* Start the replication. */
 	streamConn = pglogical_connect_replica(MySubscription->origin_if->dsn,
 										   MySubscription->name, "catchup");
@@ -1084,7 +1152,7 @@ pglogical_sync_main(Datum main_arg)
 	pglogical_identify_system(streamConn, NULL, NULL, NULL, NULL);
 
 	pglogical_start_replication(streamConn, MySubscription->slot_name,
-								origin_startpos, "all", NULL, tablename);
+								status_lsn, "all", NULL, tablename);
 
 	/* Leave it to standard apply code to do the replication. */
 	apply_work(streamConn);
@@ -1111,8 +1179,6 @@ create_local_sync_status(PGLogicalSyncStatus *sync)
 	HeapTuple	tup;
 	Datum		values[Natts_local_sync_state];
 	bool		nulls[Natts_local_sync_state];
-	NameData	nspname;
-	NameData	relname;
 
 	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
 	rel = heap_openrv(rv, RowExclusiveLock);
@@ -1123,29 +1189,24 @@ create_local_sync_status(PGLogicalSyncStatus *sync)
 
 	values[Anum_sync_kind - 1] = CharGetDatum(sync->kind);
 	values[Anum_sync_subid - 1] = ObjectIdGetDatum(sync->subid);
-	if (sync->nspname)
-	{
-		namestrcpy(&nspname, sync->nspname);
-		values[Anum_sync_nspname - 1] = NameGetDatum(&nspname);
-	}
+
+	if (sync->nspname.data[0])
+		values[Anum_sync_nspname - 1] = NameGetDatum(&sync->nspname);
 	else
 		nulls[Anum_sync_nspname - 1] = true;
-	if (sync->relname)
-	{
-		namestrcpy(&relname, sync->relname);
-		values[Anum_sync_relname - 1] = NameGetDatum(&relname);
-	}
+
+	if (sync->relname.data[0])
+		values[Anum_sync_relname - 1] = NameGetDatum(&sync->relname);
 	else
 		nulls[Anum_sync_relname - 1] = true;
+
 	values[Anum_sync_status - 1] = CharGetDatum(sync->status);
+	values[Anum_sync_statuslsn - 1] = LSNGetDatum(sync->statuslsn);
 
 	tup = heap_form_tuple(tupDesc, values, nulls);
 
 	/* Insert the tuple to the catalog. */
-	simple_heap_insert(rel, tup);
-
-	/* Update the indexes. */
-	CatalogUpdateIndexes(rel, tup);
+	CatalogTupleInsert(rel, tup);
 
 	/* Cleanup. */
 	heap_freetuple(tup);
@@ -1189,7 +1250,7 @@ syncstatus_fromtuple(HeapTuple tuple, TupleDesc desc)
 	Datum					d;
 	bool					isnull;
 
-	sync = (PGLogicalSyncStatus *) palloc(sizeof(PGLogicalSyncStatus));
+	sync = (PGLogicalSyncStatus *) palloc0(sizeof(PGLogicalSyncStatus));
 
 	d = fastgetattr(tuple, Anum_sync_kind, desc, &isnull);
 	Assert(!isnull);
@@ -1200,21 +1261,21 @@ syncstatus_fromtuple(HeapTuple tuple, TupleDesc desc)
 	sync->subid = DatumGetObjectId(d);
 
 	d = fastgetattr(tuple, Anum_sync_nspname, desc, &isnull);
-	if (isnull)
-		sync->nspname = NULL;
-	else
-		sync->nspname = pstrdup(NameStr(*DatumGetName(d)));
+	if (!isnull)
+		namestrcpy(&sync->nspname, NameStr(*DatumGetName(d)));
 
 	d = fastgetattr(tuple, Anum_sync_relname, desc, &isnull);
-	if (isnull)
-		sync->relname = NULL;
-	else
-		sync->relname = pstrdup(NameStr(*DatumGetName(d)));
+	if (!isnull)
+		namestrcpy(&sync->relname, NameStr(*DatumGetName(d)));
 
 	d = fastgetattr(tuple, Anum_sync_status, desc, &isnull);
 	Assert(!isnull);
 	sync->status = DatumGetChar(d);
 
+	d = fastgetattr(tuple, Anum_sync_statuslsn, desc, &isnull);
+	Assert(!isnull);
+	sync->statuslsn = DatumGetLSN(d);
+
 	return sync;
 }
 
@@ -1307,14 +1368,13 @@ set_subscription_sync_status(Oid subid, char status)
 
 	values[Anum_sync_status - 1] = CharGetDatum(status);
 	replaces[Anum_sync_status - 1] = true;
+	values[Anum_sync_statuslsn - 1] = LSNGetDatum(InvalidXLogRecPtr);
+	replaces[Anum_sync_statuslsn - 1] = true;
 
 	newtup = heap_modify_tuple(oldtup, tupDesc, values, nulls, replaces);
 
 	/* Update the tuple in catalog. */
-	simple_heap_update(rel, &oldtup->t_self, newtup);
-
-	/* Update the indexes. */
-	CatalogUpdateIndexes(rel, newtup);
+	CatalogTupleUpdate(rel, &oldtup->t_self, newtup);
 
 	/* Cleanup. */
 	heap_freetuple(newtup);
@@ -1442,7 +1502,7 @@ get_unsynced_tables(Oid subid)
 
 		sync = syncstatus_fromtuple(tuple, tupDesc);
 		if (sync->status != SYNC_STATUS_READY)
-			res = lappend(res, makeRangeVar(sync->nspname, sync->relname, -1));
+			res = lappend(res, sync);
 	}
 
 	systable_endscan(scan);
@@ -1454,7 +1514,7 @@ get_unsynced_tables(Oid subid)
 /* Set the sync status for a table. */
 void
 set_table_sync_status(Oid subid, const char *nspname, const char *relname,
-					  char status)
+					  char status, XLogRecPtr statuslsn)
 {
 	RangeVar	   *rv;
 	Relation		rel;
@@ -1496,14 +1556,13 @@ set_table_sync_status(Oid subid, const char *nspname, const char *relname,
 
 	values[Anum_sync_status - 1] = CharGetDatum(status);
 	replaces[Anum_sync_status - 1] = true;
+	values[Anum_sync_statuslsn - 1] = LSNGetDatum(statuslsn);
+	replaces[Anum_sync_statuslsn - 1] = true;
 
 	newtup = heap_modify_tuple(oldtup, tupDesc, values, nulls, replaces);
 
 	/* Update the tuple in catalog. */
-	simple_heap_update(rel, &oldtup->t_self, newtup);
-
-	/* Update the indexes. */
-	CatalogUpdateIndexes(rel, newtup);
+	CatalogTupleUpdate(rel, &oldtup->t_self, newtup);
 
 	/* Cleanup. */
 	heap_freetuple(newtup);
@@ -1519,10 +1578,12 @@ set_table_sync_status(Oid subid, const char *nspname, const char *relname,
  */
 bool
 wait_for_sync_status_change(Oid subid, char *nspname, char *relname,
-							char desired_state)
+							char desired_state, XLogRecPtr *lsn)
 {
 	int rc;
 
+	*lsn = InvalidXLogRecPtr;
+
 	while (!got_SIGTERM)
 	{
 		PGLogicalWorker		   *worker;
@@ -1537,6 +1598,7 @@ wait_for_sync_status_change(Oid subid, char *nspname, char *relname,
 		}
 		if (sync->status == desired_state)
 		{
+			*lsn = sync->statuslsn;
 			CommitTransactionCommand();
 			return true;
 		}
@@ -1599,13 +1661,15 @@ truncate_table(char *nspname, char *relname)
 	 * for TruncateStmt, but if we ever decide to use standard_ProcessUtility
 	 * for other utility statements, then we must take a careful relook.
 	 */
+#ifdef PGXC
 	standard_ProcessUtility((Node *)truncate,
 			sql.data, PROCESS_UTILITY_TOPLEVEL, NULL, NULL,
-#ifdef PGXC
 			false,
-#endif
 			NULL
 			);
+#else
+	ExecuteTruncate(truncate);
+#endif
 	/* release memory allocated to create SQL statement */
 	pfree(sql.data);
 
